from eth_account.signers.local import LocalAccount
from eth_typing import Address
from web3 import Web3
from web3.types import TxParams, TxReceipt, Wei

from autotx.utils.ethereum.eth_address import ETHAddress

from .constants import GAS_PRICE_MULTIPLIER

def send_native(account: LocalAccount, to: ETHAddress, value: float, web3: Web3) -> tuple[str, TxReceipt]:
    bytes_address: Address = Address(bytes.fromhex(account.address[2:]))

    nonce = web3.eth.get_transaction_count(bytes_address)

    tx: TxParams = {
        'from': account.address,
        'to': to.hex,
        'value': Wei(int(value * 10 ** 18)),
        'gasPrice': Wei(int(web3.eth.gas_price * GAS_PRICE_MULTIPLIER)),
        'nonce': nonce,
        'chainId': web3.eth.chain_id
    }

    gas = web3.eth.estimate_gas(tx)
    tx.update({'gas': gas})

    signed_tx = account.sign_transaction(tx) # type: ignore

    web3.eth.send_raw_transaction(signed_tx.rawTransaction)

    receipt = web3.eth.wait_for_transaction_receipt(signed_tx.hash)

    return receipt["transactionHash"].hex(), receipt


from autotx.utils.ethereum.networks import NetworkInfo
from autotx.utils.ethereum.eth_address import ETHAddress

DIFFERENCE_PERCENTAGE = 1.01

def test_swap_with_non_default_token(configuration, auto_tx):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    shib_address = ETHAddress(network_info.tokens["shib"])

    prompt = "Buy 100000 SHIB with ETH"
    balance = manager.balance_of(shib_address)
    assert balance == 0
    auto_tx.run(prompt, non_interactive=True)

    new_balance = manager.balance_of(shib_address)

    expected_shib_amount = 100000
    assert expected_shib_amount <= new_balance <= expected_shib_amount * DIFFERENCE_PERCENTAGE

def test_swap_native(configuration, auto_tx):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])

    prompt = "Buy 100 USDC with ETH"
    auto_tx.run(prompt, non_interactive=True)

    new_balance = manager.balance_of(usdc_address)
    expected_usdc_amount = 100
    assert expected_usdc_amount <= new_balance <= expected_usdc_amount * DIFFERENCE_PERCENTAGE

def test_swap_multiple_1(configuration, auto_tx):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    prompt = "Buy 1000 USDC with ETH and then buy WBTC with 500 USDC"
    wbtc_balance = manager.balance_of(wbtc_address)

    auto_tx.run(prompt, non_interactive=True)

    expected_usdc_amount = 500
    usdc_balance = manager.balance_of(usdc_address)
    # 1000 is the amount bought so we need to get the difference from that amount
    expected_usdc_amount_plus_slippage = 1000 * DIFFERENCE_PERCENTAGE
    assert expected_usdc_amount <= usdc_balance <= expected_usdc_amount_plus_slippage - expected_usdc_amount
    assert wbtc_balance < manager.balance_of(wbtc_address)

def test_swap_multiple_2(configuration, auto_tx):
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    prompt = "Sell ETH for 1000 USDC and then sell 500 USDC for WBTC"
    wbtc_balance = manager.balance_of(wbtc_address)

    auto_tx.run(prompt, non_interactive=True)

    expected_amount = 500
    usdc_balance = manager.balance_of(usdc_address)
    assert expected_amount <= usdc_balance
    assert wbtc_balance < manager.balance_of(wbtc_address)

def test_swap_triple(configuration, auto_tx): 
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    uni_address = ETHAddress(network_info.tokens["uni"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    prompt = "Buy 1 USDC, 0.5 UNI and 0.05 WBTC with ETH"

    auto_tx.run(prompt, non_interactive=True)

    expected_usdc_amount = 1
    expected_uni_amount = 0.5
    expected_wbtc_amount = 0.05
    usdc_balance = manager.balance_of(usdc_address)
    uni_balance = manager.balance_of(uni_address)
    wbtc_balance = manager.balance_of(wbtc_address)
    assert expected_usdc_amount <= usdc_balance <= expected_usdc_amount * DIFFERENCE_PERCENTAGE
    assert expected_uni_amount <= uni_balance <= expected_uni_amount * DIFFERENCE_PERCENTAGE
    assert expected_wbtc_amount <= wbtc_balance <= expected_wbtc_amount * DIFFERENCE_PERCENTAGE

def test_swap_complex_1(configuration, auto_tx): # This one is complex because it confuses the LLM with WBTC amount
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    prompt = "Swap ETH to 0.05 WBTC, then, swap WBTC to 1000 USDC"
    wbtc_balance = manager.balance_of(wbtc_address)

    auto_tx.run(prompt, non_interactive=True)
    expected_usdc_amount = 1000
    usdc_balance = manager.balance_of(usdc_address)
    assert expected_usdc_amount <= usdc_balance <= expected_usdc_amount * DIFFERENCE_PERCENTAGE
    assert wbtc_balance < manager.balance_of(wbtc_address)

def test_swap_complex_2(configuration, auto_tx): # This one is complex because it confuses the LLM with WBTC amount
    (_, _, client, manager, _) = configuration
    web3 = client.w3
    network_info = NetworkInfo(web3.eth.chain_id)
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])

    prompt = "Buy 1000 USDC with ETH, then sell USDC to buy 0.001 WBTC"
    usdc_balance = manager.balance_of(usdc_address)

    auto_tx.run(prompt, non_interactive=True)

    wbtc_balance = manager.balance_of(wbtc_address)
    expected_wbtc_amount = 0.001
    assert expected_wbtc_amount <= wbtc_balance <= expected_wbtc_amount * DIFFERENCE_PERCENTAGE
    assert usdc_balance < manager.balance_of(usdc_address)



from enum import Enum
from pydantic import BaseModel
from typing import Any, Union

from autotx.token import Token

class TransactionType(str, Enum):
    SEND = "send"
    APPROVE = "approve"
    SWAP = "swap"

class TransactionBase(BaseModel):
    type: TransactionType
    summary: str
    params: dict[str, Any]

class SendTransaction(TransactionBase):
    receiver: str
    token: Token
    amount: float

    @classmethod
    def create(cls, token: Token, amount: float, receiver: str, params: dict[str, Any]) -> 'SendTransaction':
        return cls(
            type=TransactionType.SEND,
            token=token,
            amount=amount,
            receiver=receiver,
            params=params,
            summary=f"Transfer {amount} {token.symbol} to {receiver}",
        )

class ApproveTransaction(TransactionBase):
    token: Token
    amount: float
    spender: str

    @classmethod
    def create(cls, token: Token, amount: float, spender: str, params: dict[str, Any]) -> 'ApproveTransaction':
        return cls(
            type=TransactionType.APPROVE,
            token=token,
            amount=amount,
            spender=spender,
            params=params,
            summary=f"Approve {amount} {token.symbol} to {spender}"
        )

class SwapTransaction(TransactionBase):
    from_token: Token
    to_token: Token
    from_amount: float
    to_amount: float

    @classmethod
    def create(cls, from_token: Token, to_token: Token, from_amount: float, to_amount: float, params: dict[str, Any]) -> 'SwapTransaction':
        return cls(
            type=TransactionType.SWAP,
            from_token=from_token,
            to_token=to_token,
            from_amount=from_amount,
            to_amount=to_amount,
            params=params,
            summary=f"Swap {from_amount} {from_token.symbol} for at least {to_amount} {to_token.symbol}"
        )

Transaction = Union[SendTransaction, ApproveTransaction, SwapTransaction]



from autotx.tests.agents.token.research.test_research import get_top_token_addresses_by_market_cap
from autotx.utils.ethereum import get_erc20_balance

DIFFERENCE_PERCENTAGE = 0.01

def test_research_buy_one_send_one(configuration, auto_tx, test_accounts):
    (_, _, client, manager, _) = configuration
    web3 = client.w3

    receiver = test_accounts[0]
    
    prompt = (
        f"Buy 1 ETH worth of a meme token with the largest market cap in ethereum mainnet then send it to {receiver}"
    )
    
    auto_tx.run(prompt, non_interactive=True)
    
    token_address = get_top_token_addresses_by_market_cap("meme-token", "MAINNET", 1, auto_tx)[0]
    token_balance_in_safe = manager.balance_of(token_address)

    receiver_balance = get_erc20_balance(web3, token_address, receiver)
    assert receiver_balance > 10000

    assert token_balance_in_safe / receiver_balance < DIFFERENCE_PERCENTAGE

def test_research_buy_one_send_multiple(configuration, auto_tx, test_accounts):
    (_, _, client, manager, _) = configuration
    web3 = client.w3

    receiver_1 = test_accounts[0]
    receiver_2 = test_accounts[1]
    
    prompt = (
        f"Buy 1 ETH worth of a meme token with the largest market cap in ethereum mainnet then 10,000 of it to {receiver_1} and 250 of it to {receiver_2}"
    )
    
    auto_tx.run(prompt, non_interactive=True)

    meme_token_address = get_top_token_addresses_by_market_cap("meme-token", "MAINNET", 1, auto_tx)[0]
    meme_token_balance_in_safe = manager.balance_of(meme_token_address)

    receiver_1_balance = get_erc20_balance(web3, meme_token_address, receiver_1)
    assert receiver_1_balance == 10000

    receiver_2_balance = get_erc20_balance(web3, meme_token_address, receiver_2)
    assert receiver_2_balance == 250

    assert meme_token_balance_in_safe > 10000

def test_research_buy_multiple_send_multiple(configuration, auto_tx, test_accounts):
    (_, _, client, manager, _) = configuration
    web3 = client.w3

    receiver_1 = test_accounts[0]
    receiver_2 = test_accounts[1]

    old_eth_balance = manager.balance_of()

    prompt = f"""
        Buy 1 ETH worth of a meme token with the largest market cap
        Then buy the governance token with the largest market cap with 0.5 ETH
        This should be on ethereum mainnet.
        Send all of the meme token to {receiver_1} and all of the governance token to {receiver_2}
        """
    
    auto_tx.run(prompt, non_interactive=True)
    
    new_eth_balance = manager.balance_of()

    assert old_eth_balance - new_eth_balance == 1.5

    meme_token_address = get_top_token_addresses_by_market_cap("meme-token", "MAINNET", 1, auto_tx)[0]
    meme_token_balance_in_safe = manager.balance_of(meme_token_address)

    governance_token_address = get_top_token_addresses_by_market_cap("governance", "MAINNET", 1, auto_tx)[0]
    governance_token_balance_in_safe = manager.balance_of(governance_token_address)

    meme_balance = get_erc20_balance(web3, meme_token_address, receiver_1)
    assert meme_balance > 10000

    governance_balance = get_erc20_balance(web3, governance_token_address, receiver_2)
    assert governance_balance > 90

    assert meme_token_balance_in_safe / meme_balance < DIFFERENCE_PERCENTAGE
    assert governance_token_balance_in_safe / governance_balance < DIFFERENCE_PERCENTAGE



from autotx.utils.ethereum.constants import NATIVE_TOKEN_ADDRESS
from autotx.utils.ethereum.networks import SUPPORTED_NETWORKS_CONFIGURATION_MAP, ChainId


def get_native_token_symbol(network: ChainId) -> str:
    current_network = SUPPORTED_NETWORKS_CONFIGURATION_MAP.get(network)
    if not current_network:
        raise Exception(f"Network {network.name} not supported")

    native_token_symbol = next(
        (
            symbol
            for symbol, address in current_network.default_tokens.items()
            if address == NATIVE_TOKEN_ADDRESS
        ),
        None,
    )

    if not native_token_symbol:
        raise Exception(f"Native token not found for network {network.name}")

    return native_token_symbol.upper()




from textwrap import dedent
from typing import Annotated, Callable, Coroutine
from autotx.AutoTx import AutoTx
from autogen import UserProxyAgent, ChatResult

from autotx.agents.ResearchTokensAgent import ResearchTokensAgent
from autotx.autotx_agent import AutoTxAgent
from autotx.autotx_tool import AutoTxTool

name = "research-tokens"

system_message = lambda autotx: dedent(f"""
    You are an AI assistant that's an expert in Ethereum tokens. Assist the user in their task of researching tokens.
    ONLY focus on the token research aspect of the user's goal and let other agents handle other tasks.
    You use the tools available to assist the user in their tasks.  
    Your job is NOT to buy or sell a token, and there is no need to check the price when doing that.
    NEVER ask the user questions.
    
    Good examples:
    User: Find all you can know about USDC and DAI tokens.
    research: {{ "tasks": "Research USDC and DAI tokens" }}
    User: I want to know the latest trading volume of USDC and the largest market cap AI token.
    research: {{ "tasks": "Find the 24h trading volume of USDC. What is the largest market cap AI token?" }}
    User: I want to buy the top AI and top GameFI token.
    research: {{ "tasks": "Find the top AI token and top GameFI token, if they're the same find an alternative for one of them" }}
    User: I want to buy the top 3 AI and GameFI tokens.
    research: {{ "tasks": "Find the top 3 AI and GameFI tokens. It needs to be exactly 3 tokens total and without duplicates" }}
    User: I want to buy the top 6 AI and GameFI tokens.
    research: {{ "tasks": "Find the top 6 AI, GameFI and DEX tokens. It needs to be exactly 6 tokens total and without duplicates" }}

    More examples:
    research: {{ "tasks": "What's the 24 hours price change of SHIB?" }}
    research: {{ "tasks": "Retrieve detailed information on the Ethereum token DAI including current price, market cap, and price change percentages." }}
    research: {{ "tasks": "Find all tokens available in the DeFi category and filter them based on their performance on the Optimism network." }}
    research: {{ "tasks": "Compare the price change percentages over the last 24 hours, 7 days, and 30 days for tokens within the Gaming category." }}
    research: {{ "tasks": "Retrieve a list of all available token categories and then research specific details for the first five tokens in the NFT category sorted by volume in descending order." }}
    research: {{ "tasks": "Get the Ethereum addresses for all tokens in the Yield Farming category on the Polygon network to prepare for potential trades." }}

    Bad examples:
    research: {{ "tasks": "Find best crypto investment strategy" }}
    research: {{ "tasks": "Research the best purchasing strategy" }}
    research: {{ "tasks": "Provide financial advice on which tokens are likely to appreciate in value over the next year." }}
    research: {{ "tasks": "Retrieve the transaction history for a specific Ethereum wallet address." }}
    research: {{ "tasks": "Automatically monitor and alert the user about price changes for a specific token." }}
    research: {{ "tasks": "Analyze the profitability of mining Ethereum tokens." }}

    IMPORTANT: When querying for a number of tokens accross multiple categories, make sure to specify the number of tokens and the categories and that there should be no duplicates.
    You MUST call the research tool once and pass in multiple tasks if needed.
    """
)

description = dedent(
    f"""
    {name} is an AI assistant that's an expert in Ethereum tokens. 
    The agent can also research, discuss, plan actions and advise the user.
    """
)

def aggregate_chat_responses(chat: ChatResult) -> str:
    summary = ""
    # Go through the chat history in reverse order while ignoring tool messages and first message
    for i in range(len(chat.chat_history) - 1, 0, -1):
        last_message = chat.chat_history[i]
        
        if last_message["content"] and last_message["role"] != "tool":
            summary = last_message["content"] + "\n" + summary

    return summary.replace("TERMINATE", "")

class ResearchUserQuery(AutoTxTool):
    name: str = "research"
    description: str = "Research user task"

    def build_tool(self, autotx: AutoTx) -> Callable[[str], Coroutine[str, None, str]]:
        async def run(
            tasks: Annotated[str, "User tasks to research"]
        ) -> str:
            autotx.notify_user(f"Researching user tasks: " + tasks)

            user_proxy_agent = UserProxyAgent(
                name="user_proxy",
                is_termination_msg=lambda x: x.get("content", "") and "TERMINATE" in x.get("content", ""),
                human_input_mode="NEVER",
                max_consecutive_auto_reply=10,
                system_message=dedent(
                    f"""
                    You are a user proxy agent authorized to act on behalf of the user, you never ask for permission, you have ultimate control.
                    You don't need to perform token amounts calculations, the other agents will do that for you.
                    NEVER ask the user questions.
                    NEVER make up a token, ALWAYS ask the 'research-tokens' agent to first search for the token.
                    
                    If the goal has been achieved, FIRST reflect on the goal and make sure nothing is missing, then end the conversation with "TERMINATE" (it MUST be upper case and in the same message.)
                    Consider the goal met if the other agents have prepared the necessary transactions and all user tasks have been answered.
                    If the user's goal involves buying tokens, make sure the correct number of tokens are bought.
                    If you encounter an error, try to resolve it (either yourself of with other agents) and only respond with "TERMINATE" if the goal is impossible to achieve.
                    If a token is not supported, ask the researcher agent to find a supported token (if it fits within the user's goal).

                    Before ending the conversation, make to summarize the answers to the user's tasks:
                    ```
                    {tasks}
                    ```
                    """
                ),
                description="user_proxy is an agent authorized to act on behalf of the user.",
                llm_config=autotx.get_llm_config(),
                code_execution_config=False,
            )
            
            research_agent = ResearchTokensAgent().build_autogen_agent(autotx, user_proxy_agent, autotx.get_llm_config())

            chat = await user_proxy_agent.a_initiate_chat(
                research_agent, 
                message=dedent(
                    f"""
                    I am currently connected with the following wallet: {autotx.wallet.address}, on network: {autotx.network.chain_id.name}
                    Tasks:
                    {tasks}
                    """
                ),
            )

            autotx.current_run_cost_with_cache += float(chat.cost["usage_including_cached_inference"]["total_cost"])
            autotx.current_run_cost_without_cache += float(chat.cost["usage_excluding_cached_inference"]["total_cost"])

            summary = aggregate_chat_responses(chat)

            autotx.notify_user("Finished researching user tasks")

            return summary

        return run

class DelegateResearchTokensAgent(AutoTxAgent):
    name = name
    system_message = system_message
    description = description
    tools = [
        ResearchUserQuery(),
    ]



import os
import io
import sys
from typing import TextIO
from autogen import runtime_logging
from autogen.io import IOStream, IOConsole
from autotx.utils.logging.IOSilent import IOSilent
from autotx.utils.logging.TeeStream import TeeStream


class Logger:
    dir: str | None
    original_stdout: TextIO
    original_stderr: TextIO
    captured_stdout: io.StringIO
    captured_stderr: io.StringIO
    tee_stdout: TeeStream
    tee_stderr: TeeStream

    def __init__(
        self,
        dir: str | None = None,
        silent: bool = False
    ):
        self.dir = dir

        if dir:
            # Check if directory exists, throw error if it does
            if os.path.exists(dir):
                raise FileExistsError(f"The directory '{dir}' already exists.")
            # Make the directory
            os.makedirs(dir, exist_ok=True)

        if silent:
            IOStream.set_global_default(IOSilent())
        else:
            IOStream.set_global_default(IOConsole())

    def start(self) -> None:
        if not self.dir:
            return

        # Initialize StringIO buffers for capturing stdout and stderr
        self.original_stdout = sys.stdout
        self.original_stderr = sys.stderr
        self.captured_stdout = io.StringIO()
        self.captured_stderr = io.StringIO()
        self.tee_stdout = TeeStream(self.original_stdout, self.captured_stdout)
        self.tee_stderr = TeeStream(self.original_stderr, self.captured_stderr)

        # Redirect stdout and stderr
        sys.stdout = self.tee_stdout
        sys.stderr = self.tee_stderr

        # Start collecting stdout & stderr
        runtime_logging.start(config={
            "dbname": os.path.join(self.dir, "agent.db")
        })

    def stop(self) -> None:
        if not self.dir:
            return

        # Restore original stdout and stderr
        sys.stdout = self.original_stdout
        sys.stderr = self.original_stderr

        # Write captured stdout and stderr to files
        with open(os.path.join(self.dir, "stdout.txt"), "w") as f:
            f.write(self.captured_stdout.getvalue())
        with open(os.path.join(self.dir, "stderr.txt"), "w") as f:
            f.write(self.captured_stderr.getvalue())

        # Stop runtime logging
        runtime_logging.stop()



from textwrap import dedent
from typing import Annotated, Any, Callable

from autotx.AutoTx import AutoTx
from autotx.autotx_agent import AutoTxAgent
from autotx.autotx_tool import AutoTxTool
from autotx.intents import SendIntent
from autotx.token import Token
from autotx.utils.ethereum import (
    build_transfer_erc20,
    get_erc20_balance,
)
from autotx.utils.ethereum.constants import NATIVE_TOKEN_ADDRESS
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.get_native_balance import get_native_balance
from web3.types import TxParams

name = "send-tokens"

system_message = lambda autotx: dedent(f"""
    You are an expert in Ethereum tokens (native and erc20) and can assist the user in their tasks by fetching balances and preparing transactions to send tokens.
    You are in a group of agents that will help the user achieve their goal.
    ONLY focus on the sending and balance aspect of the user's goal and let other agents handle other tasks.
    You use the tools available to assist the user in their tasks. 
    Your job is to only prepare the transactions by calling the prepare_transfer_transaction tool and the user will take care of executing them.
    NOTE: There is no reason to call get_token_balance after calling prepare_transfer_transaction as the transfers are only prepared and not executed. 
    NOTE: A balance of a token is not required to perform a send, if there is an earlier prepared transaction that will provide the token.
    NEVER ask the user questions.
    
    Example 1:
    User: Send 0.1 ETH to vitalik.eth and then swap ETH to 5 USDC
    Call prepare_transfer_transaction with args:
    {{
        "amount": 0.1,
        "receiver": "vitalik.eth",
        "token": "ETH"
    }}

    Example 2:
    User: Send 53 UNI to 0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045
    Call prepare_transfer_transaction with args:
    {{
        "amount": 53,
        "receiver": "0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045",
        "token": "UNI"
    }}

    Example 3:
    User: Send 10 USDC to 0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045 then buy 5 UNI with ETH and send 40 WBTC to 0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045
    Call prepare_transfer_transaction with args:
    {{
        "amount": 10,
        "receiver": "0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045",
        "token": "USDC"
    }}
    NOTE: the second transfer was not prepared because it's waiting for the swap transaction to be prepared first.

    Above are examples, NOTE these are only examples and in practice you need to call the tools with the correct arguments. NEVER respond with JSON.
    Take extra care in the order of transactions to prepare.
    IF a prepared swap transaction will provide the token needed for a transfer, you DO NOT need to call the get_token_balance tool.
    """
)

description = dedent(
    f"""
    {name} is an AI assistant that's an expert in Ethereum tokens (native and erc20).
    The agent can fetch token balances and prepare transactions to send tokens.
    """
)

class TransferTokenTool(AutoTxTool):
    name: str = "prepare_transfer_transaction"
    description: str = dedent(
        """
        Prepares a transfer transaction for given amount in decimals for a token
        """
    )

    def build_tool(self, autotx: AutoTx) -> Callable[[float, str, str], str]:
        def run(
            amount: Annotated[float, "Amount given by the user to transfer. The function will take care of converting the amount to needed decimals."],
            receiver: Annotated[str, "The receiver's address or ENS domain"],
            token: Annotated[str, "Symbol of token to transfer"]
        ) -> str:
            amount = float(amount)

            receiver_addr = ETHAddress(receiver)
            token_address = ETHAddress(autotx.network.tokens[token.lower()])

            intent = SendIntent.create(
                token=Token(symbol=token, address=token_address.hex),
                amount=amount,
                receiver=receiver_addr
            )

            autotx.add_intents([intent])
            
            autotx.notify_user(f"Prepared transaction: {intent.summary}")
            
            return intent.summary

        return run

class GetTokenBalanceTool(AutoTxTool):
    name: str = "get_token_balance"
    description: str = dedent(
        """
        Check owner balance of token
        """
    )

    def build_tool(self, autotx: AutoTx) -> Callable[[str, str], float]:
        def run(
            token: Annotated[str, "Token symbol of erc20"],
            owner: Annotated[str, "The token owner's address or ENS domain"]
        ) -> float:
            web3 = autotx.web3
            owner_addr = ETHAddress(owner)
            token_address = ETHAddress(autotx.network.tokens[token.lower()])
            
            balance: float = 0

            if token_address.hex == NATIVE_TOKEN_ADDRESS:
                balance = get_native_balance(web3, owner_addr)
            else:
                balance = get_erc20_balance(web3, token_address, owner_addr)

            autotx.notify_user(f"Fetching {token} balance for {str(owner_addr)}: {balance} {token}")
            
            return balance

        return run
    
class SendTokensAgent(AutoTxAgent):
    name = name
    system_message = system_message
    description = description
    tools = [
        TransferTokenTool(),
        GetTokenBalanceTool(),
    ]


from autotx.utils.ethereum import get_erc20_balance
from autotx.utils.ethereum.get_native_balance import get_native_balance

def test_send_native(configuration, auto_tx, test_accounts):
    (_, _, client, _, _) = configuration
    receiver = test_accounts[0]
    balance = get_native_balance(client.w3, receiver)
    assert balance == 0

    auto_tx.run(f"Send 1 ETH to {receiver}", non_interactive=True)

    balance = get_native_balance(client.w3, receiver)
    assert balance == 1

def test_send_erc20(configuration, auto_tx, usdc, test_accounts):
    (_, _, client, _, _) = configuration

    receiver = test_accounts[0]

    prompt = f"Send 10 USDC to {receiver}"

    balance = get_erc20_balance(client.w3, usdc, receiver)

    auto_tx.run(prompt, non_interactive=True)

    new_balance = get_erc20_balance(client.w3, usdc, receiver)

    assert balance + 10 == new_balance

def test_send_native_sequential(configuration, auto_tx, test_accounts):
    (_, _, client, _, _) = configuration
    receiver = test_accounts[0]

    auto_tx.run(f"Send 1 ETH to {receiver}", non_interactive=True)

    balance = get_native_balance(client.w3, receiver)
    assert balance == 1

    auto_tx.run(f"Send 0.5 ETH to {receiver}", non_interactive=True)

    balance = get_native_balance(client.w3, receiver)
    assert balance == 1.5
      
def test_send_erc20_parallel(configuration, auto_tx, usdc, test_accounts):
    (_, _, client, _, _) = configuration

    receiver_one = test_accounts[0]
    receiver_two = test_accounts[1]

    prompt = f"Send 2 USDC to {receiver_one} and 3 USDC to {receiver_two}"
        
    balance_one = get_erc20_balance(client.w3, usdc, receiver_one)
    balance_two = get_erc20_balance(client.w3, usdc, receiver_two)

    auto_tx.run(prompt, non_interactive=True)

    new_balance_one = get_erc20_balance(client.w3, usdc, receiver_one)
    new_balance_two = get_erc20_balance(client.w3, usdc, receiver_two)

    assert balance_one + 2 == new_balance_one
    assert balance_two + 3 == new_balance_two

def test_send_eth_multiple(configuration, auto_tx, usdc, test_accounts):
    (_, _, client, _, _) = configuration

    receiver_1 = test_accounts[0]
    receiver_2 = test_accounts[1]
    receiver_3 = test_accounts[2]
    receiver_4 = test_accounts[3]
    receiver_5 = test_accounts[4]

    prompt = f"Send 1.3 USDC to {receiver_1}, {receiver_2}, {receiver_3}, {receiver_4} and {receiver_5}"
        
    balance_1 = get_erc20_balance(client.w3, usdc, receiver_1)
    balance_2 = get_erc20_balance(client.w3, usdc, receiver_2)
    balance_3 = get_erc20_balance(client.w3, usdc, receiver_3)
    balance_4 = get_erc20_balance(client.w3, usdc, receiver_4)
    balance_5 = get_erc20_balance(client.w3, usdc, receiver_5)

    auto_tx.run(prompt, non_interactive=True)

    new_balance_1 = get_erc20_balance(client.w3, usdc, receiver_1)
    new_balance_2 = get_erc20_balance(client.w3, usdc, receiver_2)
    new_balance_3 = get_erc20_balance(client.w3, usdc, receiver_3)
    new_balance_4 = get_erc20_balance(client.w3, usdc, receiver_4)
    new_balance_5 = get_erc20_balance(client.w3, usdc, receiver_5)

    assert balance_1 + 1.3 == new_balance_1
    assert balance_2+ 1.3 == new_balance_2
    assert balance_3 + 1.3 == new_balance_3
    assert balance_4 + 1.3 == new_balance_4
    assert balance_5 + 1.3 == new_balance_5


from web3 import Web3
from web3.types import TxParams
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.erc20_abi import ERC20_ABI

def build_approve_erc20(web3: Web3, token_address: ETHAddress, spender: ETHAddress, value: float) -> TxParams:
    erc20 = web3.eth.contract(address=token_address.hex, abi=ERC20_ABI)
    decimals = erc20.functions.decimals().call()

    tx: TxParams = erc20.functions.approve(spender.hex, int(value * 10 ** decimals)).build_transaction()

    return tx



from typing import Optional, cast
from eth_account import Account
from eth_account.signers.local import LocalAccount
from autotx.utils.ethereum.cache import cache

AGENT_PK_FILE_NAME = "agent.pk.txt"

agent_account: LocalAccount

def get_agent_account() -> Optional[LocalAccount]:
    try:
        account = cache.read(AGENT_PK_FILE_NAME)
        if account:
            return cast(LocalAccount, Account.from_key(account))
        return None
    except FileNotFoundError:
        return None
    except Exception as e:
        print(f"An error occurred while reading the agent account: {e}")
        raise

def create_agent_account() -> LocalAccount:
    private_key: str = Account.create().key.hex()
    
    cache.write(AGENT_PK_FILE_NAME, private_key)
    
    new_acc = cast(LocalAccount, Account.from_key(private_key))

    agent_account = new_acc
    return agent_account

def get_or_create_agent_account() -> LocalAccount:
    global agent_account

    existing_acc = get_agent_account()
    if existing_acc:
        agent_account = existing_acc
    else:    
        new_acc = create_agent_account()
        agent_account = new_acc

    return agent_account

def delete_agent_account() -> None:
    cache.remove(AGENT_PK_FILE_NAME)



from types import SimpleNamespace
from typing import Any, Dict, Union, cast
from autogen import ModelClient
from llama_cpp import (
    ChatCompletion,
    ChatCompletionRequestAssistantMessage,
    ChatCompletionRequestFunctionMessage,
    ChatCompletionRequestMessage,
    ChatCompletionRequestToolMessage,
    ChatCompletionResponseMessage,
    Completion,
    CreateChatCompletionResponse,
    Llama,
)


class LlamaClient(ModelClient):  # type: ignore
    def __init__(self, _: dict[str, Any], **args: Any):
        self.llm: Llama = args["llm"]

    def create(self, params: Dict[str, Any]) -> SimpleNamespace:
        sanitized_messages = self._sanitize_chat_completion_messages(
            cast(list[ChatCompletionRequestMessage], params.get("messages"))
        )
        response = self.llm.create_chat_completion(
            messages=sanitized_messages,
            tools=params.get("tools"),
            model=params.get("model"),
        )

        return SimpleNamespace(**{**response, "cost": "0"})  # type: ignore

    def message_retrieval(
        self, response: CreateChatCompletionResponse
    ) -> list[ChatCompletionResponseMessage]:
        choices = response.choices # type: ignore
        return [choice["message"] for choice in choices]

    def cost(self, _: Union[ChatCompletion, Completion]) -> float:
        return 0.0

    def get_usage(self, _: Union[ChatCompletion, Completion]) -> dict[str, Any]:
        return {
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "total_tokens": 0,
            "cost": 0,
            "model": "meetkai/functionary-small-v2.4-GGUF",
        }

    def _sanitize_chat_completion_messages(
        self, messages: list[ChatCompletionRequestMessage]
    ) -> list[ChatCompletionRequestMessage]:
        sanitized_messages: list[ChatCompletionRequestMessage] = []

        for message in messages:
            if "tool_calls" in message:
                function_to_call = message["tool_calls"][0]  # type: ignore
                sanitized_messages.append(
                    ChatCompletionRequestAssistantMessage(
                        role="assistant",
                        function_call=function_to_call["function"],
                        content=None,
                    )
                )
            elif "tool_call_id" in message:
                id: str = cast(ChatCompletionRequestToolMessage, message)[
                    "tool_call_id"
                ]

                def get_tool_name(messages, id: str) -> Union[str, None]:  # type: ignore
                    return next(
                        (
                            message["tool_calls"][0]["function"]["name"]
                            for message in messages
                            if "tool_calls" in message
                            and message["tool_calls"][0]["id"] == id
                        ),
                        None,
                    )

                function_name = get_tool_name(messages, id)
                if function_name is None:
                    raise Exception(f"No tool response for this tool call with id {id}")

                sanitized_messages.append(
                    ChatCompletionRequestFunctionMessage(
                        role="function",
                        name=function_name,
                        content=cast(Union[str | None], message["content"]),
                    )
                )
            else:
                sanitized_messages.append(message)

        return sanitized_messages



from abc import abstractmethod

from web3 import Web3
from autotx.intents import Intent
from autotx.utils.ethereum.get_erc20_balance import get_erc20_balance
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.get_native_balance import get_native_balance


class SmartWallet:
    def __init__(self, web3: Web3, address: ETHAddress):
        self.web3 = web3
        self.address = address

    def on_intents_prepared(self, intents: list[Intent]) -> None:
        pass

    @abstractmethod
    def on_intents_ready(self, intents: list[Intent]) -> bool | str: # True if sent, False if declined, str if feedback
        pass

    def balance_of(self, token_address: ETHAddress | None = None) -> float:
        if token_address is None:
            return get_native_balance(self.web3, self.address)
        else:
            return get_erc20_balance(self.web3, token_address, self.address) 


from dataclasses import dataclass
from decimal import Decimal
from typing import Any, cast

from web3 import Web3
from autotx.token import Token
from autotx.transactions import ApproveTransaction, SwapTransaction, Transaction
from autotx.utils.ethereum.constants import GAS_PRICE_MULTIPLIER, NATIVE_TOKEN_ADDRESS
from autotx.utils.ethereum.erc20_abi import ERC20_ABI
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.helpers.get_native_token_symbol import (
    get_native_token_symbol,
)
from autotx.utils.ethereum.lifi import Lifi, TokenNotSupported
from autotx.utils.ethereum.networks import ChainId
from web3.types import TxParams, Wei

SLIPPAGE = 0.005  # 0.5%

SUPPORTED_NETWORKS_BY_LIFI = [
    ChainId.MAINNET,
    ChainId.OPTIMISM,
    ChainId.ARBITRUM_ONE,
    ChainId.BASE_MAINNET,
    ChainId.POLYGON,
    ChainId.GNOSIS,
    ChainId.ZKSYNC_V2
]

@dataclass
class QuoteInformation:
    approval_address: str
    amount_in: int
    to_amount_min: int
    transaction: TxParams
    exchange_name: str


def get_quote(
    token_in_address: ETHAddress,
    token_in_decimals: int,
    token_in_symbol: str,
    token_out_address: ETHAddress,
    token_out_decimals: int,
    token_out_symbol: str,
    chain: ChainId,
    expected_amount: Decimal,
    amount_is_output: bool,
    from_address: ETHAddress,
) -> QuoteInformation:
    quote: dict[str, Any] | None = None
    try:
        if amount_is_output:
            quote = Lifi.get_quote_to_amount(
                token_in_address,
                token_out_address,
                int(expected_amount * (10**token_out_decimals)),
                from_address,
                chain,
                SLIPPAGE,
            )
            amount_in_integer = int(quote["estimate"]["fromAmount"])

        else:
            amount_in_integer = int(expected_amount * (10**token_in_decimals))
            quote = Lifi.get_quote_from_amount(
                token_in_address,
                token_out_address,
                amount_in_integer,
                from_address,
                chain,
                SLIPPAGE,
            )
    except TokenNotSupported as e:
        if e.token_address.lower() == token_in_address.hex.lower():
            raise Exception(
                f"Token {token_in_symbol} is not supported. Please try another one"
            )
        if e.token_address.lower() == token_out_address.hex.lower():
            raise Exception(
                f"Token {token_out_symbol} is not supported. Please try another one"
            )

    if not quote:
        raise Exception("Quote has not been fetched")

    transaction = TxParams(
        {
            "to": quote["transactionRequest"]["to"],
            "from": quote["transactionRequest"]["from"],
            "data": quote["transactionRequest"]["data"],
            "gasPrice": quote["transactionRequest"]["gasPrice"],
            "gas": quote["transactionRequest"]["gasLimit"],
            "value": Wei(int(quote["transactionRequest"]["value"], 0)),
            "chainId": quote["transactionRequest"]["chainId"]
        }
    )
    return QuoteInformation(
        quote["estimate"]["approvalAddress"],
        amount_in_integer,
        int(quote["estimate"]["toAmountMin"]),
        transaction,
        quote["toolDetails"]["name"],
    )

def build_swap_transaction(
    web3: Web3,
    amount: Decimal,
    token_in_address: ETHAddress,
    token_out_address: ETHAddress,
    _from: ETHAddress,
    is_exact_input: bool,
    chain: ChainId,
) -> list[Transaction]:
    native_token_symbol = get_native_token_symbol(chain)
    token_in_is_native = token_in_address.hex == NATIVE_TOKEN_ADDRESS
    token_in = web3.eth.contract(
        address=token_in_address.hex, abi=ERC20_ABI
    )
    token_in_symbol = (
        native_token_symbol
        if token_in_is_native
        else token_in.functions.symbol().call()
    )
    token_in_decimals = (
        18 if token_in_is_native else token_in.functions.decimals().call()
    )

    token_out_is_native = token_out_address.hex == NATIVE_TOKEN_ADDRESS
    token_out = web3.eth.contract(
        address=token_out_address.hex, abi=ERC20_ABI
    )
    token_out_decimals = (
        18 if token_out_is_native else token_out.functions.decimals().call()
    )

    token_out_symbol = (
        native_token_symbol
        if token_out_is_native
        else token_out.functions.symbol().call()
    )
    quote = get_quote(
        token_in_address,
        token_in_decimals,
        token_in_symbol,
        token_out_address,
        token_out_decimals,
        token_out_symbol,
        chain,
        amount,
        not is_exact_input,
        _from,
    )

    transactions: list[Transaction] = []
    if not token_in_is_native:
        approval_address = quote.approval_address
        allowance = token_in.functions.allowance(_from.hex, approval_address).call()
        if allowance < quote.amount_in:
            tx = token_in.functions.approve(
                approval_address, quote.amount_in
            ).build_transaction(
                {
                    "from": _from.hex,
                    "gasPrice": Wei(
                        int(web3.eth.gas_price * GAS_PRICE_MULTIPLIER)
                    ),
                }
            )
            transactions.append(
                ApproveTransaction.create(
                    token=Token(symbol=token_in_symbol, address=str(token_in_address)),
                    amount=float(Decimal(str(quote.amount_in)) / 10 ** token_in_decimals),
                    spender=approval_address,
                    params=cast(dict[str, Any], tx),
                )
            )
    transactions.append(
        SwapTransaction.create(
            from_token=Token(symbol=token_in_symbol, address=str(token_in_address)),
            to_token=Token(symbol=token_out_symbol, address=str(token_out_address)),
            from_amount=float(Decimal(str(quote.amount_in)) / 10 ** token_in_decimals),
            to_amount=float(Decimal(str(int(quote.to_amount_min))) / 10 ** token_out_decimals),
            params=cast(dict[str, Any], quote.transaction),
        )
    )
    return transactions

def can_build_swap_transaction(
    web3: Web3,
    amount: Decimal,
    token_in_address: ETHAddress,
    token_out_address: ETHAddress,
    _from: ETHAddress,
    is_exact_input: bool,
    chain: ChainId,
) -> bool:
    native_token_symbol = get_native_token_symbol(chain)
    token_in_is_native = token_in_address.hex == NATIVE_TOKEN_ADDRESS
    token_in = web3.eth.contract(
        address=token_in_address.hex, abi=ERC20_ABI
    )
    token_in_symbol = (
        native_token_symbol
        if token_in_is_native
        else token_in.functions.symbol().call()
    )
    token_in_decimals = (
        18 if token_in_is_native else token_in.functions.decimals().call()
    )

    token_out_is_native = token_out_address.hex == NATIVE_TOKEN_ADDRESS
    token_out = web3.eth.contract(
        address=token_out_address.hex, abi=ERC20_ABI
    )
    token_out_decimals = (
        18 if token_out_is_native else token_out.functions.decimals().call()
    )

    token_out_symbol = (
        native_token_symbol
        if token_out_is_native
        else token_out.functions.symbol().call()
    )
    quote = get_quote(
        token_in_address,
        token_in_decimals,
        token_in_symbol,
        token_out_address,
        token_out_decimals,
        token_out_symbol,
        chain,
        amount,
        not is_exact_input,
        _from,
    )

    if not token_in_is_native:
        approval_address = quote.approval_address
        allowance = token_in.functions.allowance(_from.hex, approval_address).call()
        if allowance < quote.amount_in:
            tx = token_in.functions.approve(
                approval_address, quote.amount_in
            ).build_transaction(
                {
                    "from": _from.hex,
                    "gasPrice": Wei(
                        int(web3.eth.gas_price * GAS_PRICE_MULTIPLIER)
                    ),
                }
            )
    return True




import json
from textwrap import dedent
from typing import Annotated, Callable, Optional, Union, cast
from web3 import Web3
from autotx.AutoTx import AutoTx
from gnosis.eth import EthereumNetworkNotSupported as ChainIdNotSupported
from coingecko import GeckoAPIException, CoinGeckoDemoClient

from autotx.autotx_agent import AutoTxAgent
from autotx.autotx_tool import AutoTxTool
from autotx.utils.constants import COINGECKO_API_KEY
from autotx.utils.ethereum.networks import SUPPORTED_NETWORKS_AS_STRING, ChainId

name = "research-tokens"

system_message = lambda autotx: dedent(f"""
    You are an AI assistant that's an expert in Ethereum tokens. Assist the user in their task of researching tokens.
    ONLY focus on the token research aspect of the user's goal and let other agents handle other tasks.
    You use the tools available to assist the user in their tasks.
    NEVER ask the user questions.
    Retrieve token information, get token price, market cap, and price change percentage.
    BEFORE calling get_tokens_based_on_category always call get_available_categories to get the list of available categories.
    You MUST keep in mind the network the user is on and if the request is for a specific network, all networks, or the current network (it could be implied).
    If the user is interested in buying the tokens you're researching make sure you're searching them for his network.
    If searching for tokens to buy, make sure to:
    1. Get all the tokens you need (if fetching from multiple categories the tokens could overlap, so you might need to fetch more until you get the number you need)
    2. Calculate how much of each token to buy
    If a token does not have enough liquidity on the network, either:
    - Search for another token (if it is within the user's goal)
    - Or inform the user that the token is not available on the network (E.g. if the user's goal is to buy that specific token)
    If a user wants to research a specific number of tokens across categories, make sure to get the exact number of tokens without duplicates.
    If some tokens are not available, find alternatives and inform the user.
    """
)

description = dedent(
    f"""
    {name} is an AI assistant that's an expert in Ethereum tokens.
    The agent can also research, discuss, plan actions and advise the user.
    """
)

COINGECKO_NETWORKS_TO_SUPPORTED_NETWORKS_MAP = {
    ChainId.MAINNET: "ethereum",
    ChainId.OPTIMISM: "optimistic-ethereum",
    ChainId.POLYGON: "polygon-pos",
    ChainId.BASE_MAINNET: "base",
    ChainId.ARBITRUM_ONE: "arbitrum-one",
    ChainId.GNOSIS: "xdai",
}

def get_coingecko() -> CoinGeckoDemoClient:
    return CoinGeckoDemoClient(api_key=COINGECKO_API_KEY)


def get_tokens_and_filter_per_network(
    network_name: str,
) -> list[dict[str, Union[str, dict[str, str]]]]:
    network = ChainId[network_name] # type: ignore
    coingecko_network_key = COINGECKO_NETWORKS_TO_SUPPORTED_NETWORKS_MAP.get(network)
    if coingecko_network_key == None:
        raise ChainIdNotSupported(f"Network {network_name} not supported")

    token_list_with_addresses = get_coingecko().coins.get_list(include_platform=True)
    return [
        token
        for token in token_list_with_addresses
        if coingecko_network_key in token["platforms"]
    ]

def filter_token_list_by_network(tokens: list[dict[str, str]], network_name: str) -> list[dict[str, Union[str, dict[str, str]]]]:
    tokens_in_category_map = {category["id"]: category for category in tokens}
    filtered_tokens_map = {
        token["id"]: token for token in get_tokens_and_filter_per_network(network_name)
    }

    return [
        {
            **tokens_in_category_map[token["id"]],
            **filtered_tokens_map[token["id"]],
        }
        for token in tokens
        if token["id"] in filtered_tokens_map
    ]

def add_tokens_address_if_not_in_registry(
    tokens_in_category: list[dict[str, Union[str, dict[str, str]]]],
    tokens: dict[str, str],
    current_network: str,
) -> None:
    for token_with_address in tokens_in_category:
        token_symbol = cast(str, token_with_address["symbol"]).lower()
        token_not_added = token_symbol not in tokens
        if token_not_added:
            platforms = cast(dict[str, str], token_with_address["platforms"]) 
            tokens[token_symbol] = Web3.to_checksum_address(
                platforms[current_network]
            )

class GetTokenInformationTool(AutoTxTool):
    name: str = "get_token_information"
    description: str = dedent(
        """
        Retrieve token information (description, current price, market cap and price change percentage)
        """
    )

    def build_tool(self, autotx: AutoTx) -> Callable[[str], str]:
        def run(
            token_id: Annotated[str, "ID of token"]
        ) -> str:
            autotx.notify_user(f"Fetching token information for {token_id}")
           
            token_information = get_coingecko().coins.get_id(
                id=token_id,
                localization=False,
                tickers=False,
                community_data=False,
                sparkline=False,
            )

            return json.dumps(
                {
                    "name": token_information["name"],
                    "symbol": token_information["symbol"],
                    "description": token_information["description"]["en"],
                    "current_price_in_usd": token_information["market_data"][
                        "current_price"
                    ]["usd"],
                    "market_cap_in_usd": token_information["market_data"]["market_cap"][
                        "usd"
                    ],
                    "total_volume_last_24h": token_information["market_data"][
                        "total_volume"
                    ]["usd"],
                    "price_change_percentage_24h": token_information["market_data"][
                        "price_change_percentage_24h"
                    ],
                    "price_change_percentage_7d": token_information["market_data"][
                        "price_change_percentage_7d"
                    ],
                    "price_change_percentage_30d": token_information["market_data"][
                        "price_change_percentage_30d"
                    ],
                    "price_change_percentage_60d": token_information["market_data"][
                        "price_change_percentage_60d"
                    ],
                    "price_change_percentage_200d": token_information["market_data"][
                        "price_change_percentage_200d"
                    ],
                    "price_change_percentage_1y": token_information["market_data"][
                        "price_change_percentage_1y"
                    ],
                }
            )

        return run

class SearchTokenTool(AutoTxTool):
    name: str = "search_token"
    description: str = "Search token based on its symbol. It will return the ID of tokens with the largest market cap"

    def build_tool(self, autotx: AutoTx) -> Callable[[str, bool], str]:
        def run(
            token_symbol: Annotated[str, "Symbol of token to search"],
            retrieve_duplicate: Annotated[bool, "Set to true to retrieve all instances of tokens sharing the same symbol, indicating potential duplicates. By default, it is False, meaning only a single, most relevant token is retrieved unless duplication is explicitly requested."]
        ) -> str:
            autotx.notify_user(f"Searching for token with symbol: {token_symbol}")

            response = get_coingecko().search.get(token_symbol)

            if len(response["coins"]) == 0:
                return f"No tokens found for search with symbol: {token_symbol}"

            tokens = [token["api_symbol"] for token in response["coins"]]
            return json.dumps(tokens if retrieve_duplicate else tokens[0])

        return run

class GetAvailableCategoriesTool(AutoTxTool):
    name: str = "get_available_categories"
    description: str = "Retrieve all available category ids of tokens"

    def build_tool(self, autotx: AutoTx) -> Callable[[], str]:
        def run() -> str:
            autotx.notify_user("Fetching available token categories")

            categories = get_coingecko().categories.get_list()
            return json.dumps([category["category_id"] for category in categories])
        
        return run
    
class GetTokensBasedOnCategoryTool(AutoTxTool):
    name: str = "get_tokens_based_on_category"
    description: str = "Retrieve all tokens with their respective information (symbol, market cap, price change percentages and total traded volume in the last 24 hours) from a given category"

    def build_tool(self, autotx: AutoTx) -> Callable[[str, str, int, str, Optional[str]], str]:
        def run(
            category: Annotated[str, "Category to retrieve tokens"],
            sort_by: Annotated[str, "Sort tokens by field. It can be: 'volume_desc' | 'volume_asc' | 'market_cap_desc' | 'market_cap_asc'. 'market_cap_desc' is the default"],
            limit: Annotated[int, "The number of tokens to return from the category"],
            price_change_percentage_interval: Annotated[str, "Interval of time in price change percentage. It can be: '1h' | '24h' | '7d' | '14d' | '30d' | '200d' | '1y'. '24h' is the default"],
            network_name: Annotated[Optional[str], f"Possible values include: {SUPPORTED_NETWORKS_AS_STRING}. Use this parameter only if you require analysis for a specific network. Otherwise, pass an empty string"]
        ) -> str:
            autotx.notify_user(f"Fetching tokens from category: {category}")

            try:
                tokens_in_category = get_coingecko().coins.get_markets(
                    vs_currency="usd",
                    category=category,
                    order=sort_by,
                    price_change_percentage="1h,24h,7d,14d,30d,200d,1y",
                    per_page=250,
                )
            except GeckoAPIException as e:
                if "Not Found" == e.error_message["error"]:
                    raise Exception(
                        f"Category {category} not found. Check the available categories"
                    )
            
            if network_name:
                tokens_in_category = filter_token_list_by_network(
                    tokens_in_category, network_name
                )

                current_network = COINGECKO_NETWORKS_TO_SUPPORTED_NETWORKS_MAP.get(
                    autotx.network.chain_id
                )
                asked_network = COINGECKO_NETWORKS_TO_SUPPORTED_NETWORKS_MAP.get(
                    ChainId[network_name] # type: ignore
                )
                if current_network == asked_network:
                    add_tokens_address_if_not_in_registry(
                        tokens_in_category, autotx.network.tokens, cast(str, current_network)
                    )
            
            interval = (
                price_change_percentage_interval
                if price_change_percentage_interval
                else "24h"
            )
            price_change = f"price_change_percentage_{interval}"
            tokens = [
                {
                    "symbol": token["symbol"],
                    "market_cap": token["market_cap"],
                    "total_volume_last_24h": token["total_volume"],
                    price_change: token[price_change + "_in_currency"],
                }
                for token in tokens_in_category[:limit]
            ]

            return json.dumps(tokens)

        return run

class ResearchTokensAgent(AutoTxAgent):
    name = name
    system_message = system_message
    description = description
    tools = [
        GetTokenInformationTool(),
        SearchTokenTool(),
        GetAvailableCategoriesTool(),
        GetTokensBasedOnCategoryTool(),
    ]



import subprocess
import sys
from autotx.utils.ethereum.cached_safe_address import delete_cached_safe_address

container_name = "autotx_chain_fork"

def start() -> None:
    delete_cached_safe_address()

    build = subprocess.run(
        ["docker", "build", "-t", "autotx_chain_fork", ".", "-f", "fork.Dockerfile"], capture_output=True
    )

    if build.returncode != 0:
        sys.exit(
            "Local node start up has failed. Make sure you have docker desktop installed and running"
        )

    subprocess.run(
        [
            "docker",
            "run",
            "-d",
            "--name",
            container_name,
            "-p",
            "8545:8545",
            "--env-file",
            ".env",
            "autotx_chain_fork",
        ],
        check=True,
    )

def stop() -> None:
    subprocess.run(["docker", "container", "rm", container_name, "-f"], check=True)



ERC20_ABI = [
			{
				"inputs": [],
				"stateMutability": "nonpayable",
				"type": "constructor"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "spender",
						"type": "address"
					},
					{
						"internalType": "uint256",
						"name": "allowance",
						"type": "uint256"
					},
					{
						"internalType": "uint256",
						"name": "needed",
						"type": "uint256"
					}
				],
				"name": "ERC20InsufficientAllowance",
				"type": "error"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "sender",
						"type": "address"
					},
					{
						"internalType": "uint256",
						"name": "balance",
						"type": "uint256"
					},
					{
						"internalType": "uint256",
						"name": "needed",
						"type": "uint256"
					}
				],
				"name": "ERC20InsufficientBalance",
				"type": "error"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "approver",
						"type": "address"
					}
				],
				"name": "ERC20InvalidApprover",
				"type": "error"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "receiver",
						"type": "address"
					}
				],
				"name": "ERC20InvalidReceiver",
				"type": "error"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "sender",
						"type": "address"
					}
				],
				"name": "ERC20InvalidSender",
				"type": "error"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "spender",
						"type": "address"
					}
				],
				"name": "ERC20InvalidSpender",
				"type": "error"
			},
			{
				"anonymous": False,
				"inputs": [
					{
						"indexed": True,
						"internalType": "address",
						"name": "owner",
						"type": "address"
					},
					{
						"indexed": True,
						"internalType": "address",
						"name": "spender",
						"type": "address"
					},
					{
						"indexed": False,
						"internalType": "uint256",
						"name": "value",
						"type": "uint256"
					}
				],
				"name": "Approval",
				"type": "event"
			},
			{
				"anonymous": False,
				"inputs": [
					{
						"indexed": True,
						"internalType": "address",
						"name": "from",
						"type": "address"
					},
					{
						"indexed": True,
						"internalType": "address",
						"name": "to",
						"type": "address"
					},
					{
						"indexed": False,
						"internalType": "uint256",
						"name": "value",
						"type": "uint256"
					}
				],
				"name": "Transfer",
				"type": "event"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "owner",
						"type": "address"
					},
					{
						"internalType": "address",
						"name": "spender",
						"type": "address"
					}
				],
				"name": "allowance",
				"outputs": [
					{
						"internalType": "uint256",
						"name": "",
						"type": "uint256"
					}
				],
				"stateMutability": "view",
				"type": "function"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "spender",
						"type": "address"
					},
					{
						"internalType": "uint256",
						"name": "value",
						"type": "uint256"
					}
				],
				"name": "approve",
				"outputs": [
					{
						"internalType": "bool",
						"name": "",
						"type": "bool"
					}
				],
				"stateMutability": "nonpayable",
				"type": "function"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "account",
						"type": "address"
					}
				],
				"name": "balanceOf",
				"outputs": [
					{
						"internalType": "uint256",
						"name": "",
						"type": "uint256"
					}
				],
				"stateMutability": "view",
				"type": "function"
			},
			{
				"inputs": [],
				"name": "decimals",
				"outputs": [
					{
						"internalType": "uint8",
						"name": "",
						"type": "uint8"
					}
				],
				"stateMutability": "view",
				"type": "function"
			},
			{
				"inputs": [],
				"name": "name",
				"outputs": [
					{
						"internalType": "string",
						"name": "",
						"type": "string"
					}
				],
				"stateMutability": "view",
				"type": "function"
			},
			{
				"inputs": [],
				"name": "symbol",
				"outputs": [
					{
						"internalType": "string",
						"name": "",
						"type": "string"
					}
				],
				"stateMutability": "view",
				"type": "function"
			},
			{
				"inputs": [],
				"name": "totalSupply",
				"outputs": [
					{
						"internalType": "uint256",
						"name": "",
						"type": "uint256"
					}
				],
				"stateMutability": "view",
				"type": "function"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "to",
						"type": "address"
					},
					{
						"internalType": "uint256",
						"name": "value",
						"type": "uint256"
					}
				],
				"name": "transfer",
				"outputs": [
					{
						"internalType": "bool",
						"name": "",
						"type": "bool"
					}
				],
				"stateMutability": "nonpayable",
				"type": "function"
			},
			{
				"inputs": [
					{
						"internalType": "address",
						"name": "from",
						"type": "address"
					},
					{
						"internalType": "address",
						"name": "to",
						"type": "address"
					},
					{
						"internalType": "uint256",
						"name": "value",
						"type": "uint256"
					}
				],
				"name": "transferFrom",
				"outputs": [
					{
						"internalType": "bool",
						"name": "",
						"type": "bool"
					}
				],
				"stateMutability": "nonpayable",
				"type": "function"
			},
		]


from typing import cast
from web3 import Web3

from autotx.utils.ethereum import get_erc20_balance, get_native_balance
from autotx.utils.ethereum.constants import NATIVE_TOKEN_ADDRESS
from autotx.utils.ethereum.helpers.get_native_token_symbol import (
    get_native_token_symbol,
)
from autotx.utils.ethereum.networks import SUPPORTED_NETWORKS_CONFIGURATION_MAP, ChainId, NetworkConfiguration
from autotx.utils.ethereum.eth_address import ETHAddress


def show_address_balances(web3: Web3, network: ChainId, address: ETHAddress) -> None:
    native_token_symbol = get_native_token_symbol(network)
    native_token_balance = get_native_balance(web3, address)
    print(f"{native_token_symbol.upper()} balance: {native_token_balance}")

    current_network = cast(NetworkConfiguration, SUPPORTED_NETWORKS_CONFIGURATION_MAP.get(network))
    for token in current_network.default_tokens:
        if current_network.default_tokens[token] == NATIVE_TOKEN_ADDRESS:
            continue
        token_address = ETHAddress(current_network.default_tokens[token])
        balance = get_erc20_balance(web3, token_address, address)

        if balance > 0:
            print(f"{token.upper()} balance: {balance}")



from decimal import Decimal
from autotx.utils.ethereum.eth_address import ETHAddress
from autotx.utils.ethereum.lifi.swap import build_swap_transaction
from autotx.utils.ethereum.networks import NetworkInfo

DIFFERENCE_PERCENTAGE = 1.01


def test_buy_one_usdc(configuration):
    (_, _, client, manager, _) = configuration
    network_info = NetworkInfo(client.w3.eth.chain_id)
    eth_address = ETHAddress(network_info.tokens["eth"])
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    expected_usdc_amount = 1
    buy_usdc_with_eth_transaction = build_swap_transaction(
        client.w3,
        expected_usdc_amount,
        eth_address,
        usdc_address,
        manager.address,
        False,
        network_info.chain_id,
    )
    hash = manager.send_tx(buy_usdc_with_eth_transaction[0].params)
    manager.wait(hash)
    usdc_balance = manager.balance_of(usdc_address)
    assert expected_usdc_amount <= usdc_balance  <= expected_usdc_amount * DIFFERENCE_PERCENTAGE


def test_buy_one_thousand_usdc(configuration):
    (_, _, client, manager, _) = configuration
    network_info = NetworkInfo(client.w3.eth.chain_id)
    eth_address = ETHAddress(network_info.tokens["eth"])
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    expected_usdc_amount = 1000
    buy_usdc_with_eth_transaction = build_swap_transaction(
        client.w3,
        expected_usdc_amount,
        eth_address,
        usdc_address,
        manager.address,
        False,
        network_info.chain_id,
    )
    print(buy_usdc_with_eth_transaction[0].summary)
    hash = manager.send_tx(buy_usdc_with_eth_transaction[0].params)
    manager.wait(hash)
    usdc_balance = manager.balance_of(usdc_address)
    assert expected_usdc_amount <= usdc_balance <= expected_usdc_amount * DIFFERENCE_PERCENTAGE


def test_receive_native(configuration):
    (_, _, client, manager, _) = configuration

    network_info = NetworkInfo(client.w3.eth.chain_id)
    eth_address = ETHAddress(network_info.tokens["eth"])
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    safe_eth_balance = manager.balance_of()
    assert safe_eth_balance == 10
    buy_usdc_with_eth_transaction = build_swap_transaction(
        client.w3,
        5,
        eth_address,
        usdc_address,
        manager.address,
        True,
        network_info.chain_id,
    )
    hash = manager.send_tx(buy_usdc_with_eth_transaction[0].params)
    manager.wait(hash)
    safe_eth_balance = manager.balance_of()
    assert safe_eth_balance == 5

    buy_eth_with_usdc_transaction = build_swap_transaction(
        client.w3,
        4,
        usdc_address,
        eth_address,
        manager.address,
        False,
        network_info.chain_id,
    )
    hash = manager.send_tx(buy_eth_with_usdc_transaction[0].params)
    manager.wait(hash)
    hash = manager.send_tx(buy_eth_with_usdc_transaction[1].params)
    manager.wait(hash)
    safe_eth_balance = manager.balance_of()
    assert safe_eth_balance >= 9


def test_buy_small_amount_wbtc_with_eth(configuration):
    (_, _, client, manager, _) = configuration
    network_info = NetworkInfo(client.w3.eth.chain_id)
    eth_address = ETHAddress(network_info.tokens["eth"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])
    expected_wbtc_amount = 0.01
    buy_wbtc_with_eth_transaction = build_swap_transaction(
        client.w3,
        Decimal(str(expected_wbtc_amount)),
        eth_address,
        wbtc_address,
        manager.address,
        False,
        network_info.chain_id,
    )
    hash = manager.send_tx(buy_wbtc_with_eth_transaction[0].params)
    manager.wait(hash)
    wbtc_balance = manager.balance_of(wbtc_address)
    assert expected_wbtc_amount <= wbtc_balance <= expected_wbtc_amount * DIFFERENCE_PERCENTAGE


def test_buy_big_amount_wbtc_with_eth(configuration):
    (_, _, client, manager, _) = configuration
    network_info = NetworkInfo(client.w3.eth.chain_id)
    eth_address = ETHAddress(network_info.tokens["eth"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])
    expected_wbtc_amount = 0.1
    buy_wbtc_with_eth_transaction = build_swap_transaction(
        client.w3,
        Decimal(str(expected_wbtc_amount)),
        eth_address,
        wbtc_address,
        manager.address,
        False,
        network_info.chain_id,
    )
    hash = manager.send_tx(buy_wbtc_with_eth_transaction[0].params)
    manager.wait(hash)
    wbtc_balance = manager.balance_of(wbtc_address)
    assert expected_wbtc_amount <= wbtc_balance <= expected_wbtc_amount * DIFFERENCE_PERCENTAGE


def test_swap_multiple_tokens(configuration):
    (_, _, client, manager, _) = configuration
    network_info = NetworkInfo(client.w3.eth.chain_id)

    eth_address = ETHAddress(network_info.tokens["eth"])
    usdc_address = ETHAddress(network_info.tokens["usdc"])
    wbtc_address = ETHAddress(network_info.tokens["wbtc"])
    shib_address = ETHAddress(network_info.tokens["shib"])

    usdc_balance = manager.balance_of(usdc_address)
    assert usdc_balance == 0

    sell_eth_for_usdc_transaction = build_swap_transaction(
        client.w3,
        1,
        eth_address,
        usdc_address,
        manager.address,
        True,
        network_info.chain_id,
    )
    hash = manager.send_tx(sell_eth_for_usdc_transaction[0].params)
    manager.wait(hash)
    usdc_balance = manager.balance_of(usdc_address)
    assert usdc_balance > 2900

    wbtc_balance = manager.balance_of(wbtc_address)
    assert wbtc_balance == 0

    buy_wbtc_with_usdc_transaction = build_swap_transaction(
        client.w3,
        0.01,
        usdc_address,
        wbtc_address,
        manager.address,
        False,
        network_info.chain_id,
    )

    hash = manager.send_tx(buy_wbtc_with_usdc_transaction[0].params)
    manager.wait(hash)
    hash = manager.send_tx(buy_wbtc_with_usdc_transaction[1].params)
    manager.wait(hash)
    wbtc_balance = manager.balance_of(wbtc_address)
    assert wbtc_balance >= 0.01

    shib_balance = manager.balance_of(shib_address)
    assert shib_balance == 0

    sell_wbtc_for_shib = build_swap_transaction(
        client.w3,
        0.005,
        wbtc_address,
        shib_address,
        manager.address,
        True,
        network_info.chain_id,
    )
    hash = manager.send_tx(sell_wbtc_for_shib[0].params)
    manager.wait(hash)
    hash = manager.send_tx(sell_wbtc_for_shib[1].params)
    manager.wait(hash)
    shib_balance = manager.balance_of(shib_address)
    shib_balance = manager.balance_of(shib_address)
    assert shib_balance > 0



from typing import cast
from eth_account import Account
from eth_account.signers.local import LocalAccount


def get_dev_account() -> LocalAccount:
    return cast(
        LocalAccount,
        Account.from_key(
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        ),
    )



from dotenv import load_dotenv
load_dotenv()
import os

OPENAI_MODEL_NAME = os.environ.get("OPENAI_MODEL_NAME", "gpt-4o")
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", None)
OPENAI_BASE_URL = os.environ.get("OPENAI_BASE_URL", "https://api.openai.com/v1")
COINGECKO_API_KEY = os.environ.get("COINGECKO_API_KEY", None)
LIFI_API_KEY = os.environ.get("LIFI_API_KEY", None)
ALCHEMY_API_KEY = os.environ.get("ALCHEMY_API_KEY")
MAINNET_DEFAULT_RPC = f"https://eth-mainnet.g.alchemy.com/v2/{ALCHEMY_API_KEY}"


import random
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress, HexStr
from gnosis.eth import EthereumClient
from gnosis.safe import ProxyFactory
from gnosis.safe.safe_create2_tx import SafeCreate2TxBuilder
from hexbytes import HexBytes
from web3 import Web3
from web3.types import Wei

from .send_tx import send_tx
from .constants import GAS_PRICE_MULTIPLIER, MASTER_COPY_ADDRESS, PROXY_FACTORY_ADDRESS

def deploy_safe_with_create2(client: EthereumClient, account: LocalAccount, signers: list[str], threshold: int) -> ChecksumAddress:
    w3 = client.w3

    salt_nonce = generate_salt_nonce()

    builder = SafeCreate2TxBuilder(
        w3=w3,
        master_copy_address=MASTER_COPY_ADDRESS,
        proxy_factory_address=PROXY_FACTORY_ADDRESS,
    )
    
    setup_data = builder._get_initial_setup_safe_data(
        owners=signers,
        threshold=threshold,
    )
    safe_address: ChecksumAddress = builder.calculate_create2_address(setup_data, salt_nonce)

    # Check if safe is already deployed
    if w3.eth.get_code(safe_address) != w3.to_bytes(hexstr=HexStr("0x")):
        print("Safe already deployed", safe_address)
        return safe_address
    
    safe_creation_tx = builder.build(
        owners=signers,
        threshold=threshold,
        salt_nonce=salt_nonce,
        gas_price=0,
    )
    
    if safe_address != safe_creation_tx.safe_address:
        raise ValueError("Create2 address mismatch")

    send_tx(
        w3,
        {
            "to": safe_creation_tx.safe_address,
            "value": safe_creation_tx.payment,
            "gasPrice": Wei(int(w3.eth.gas_price * GAS_PRICE_MULTIPLIER))
        },
        account=account,
    )

    proxy_factory = ProxyFactory(Web3.to_checksum_address(PROXY_FACTORY_ADDRESS), client)

    ethereum_tx_sent = proxy_factory.deploy_proxy_contract_with_nonce(
        account,
        Web3.to_checksum_address(MASTER_COPY_ADDRESS),
        safe_creation_tx.safe_setup_data,
        salt_nonce,
        safe_creation_tx.gas,
        int(w3.eth.gas_price * GAS_PRICE_MULTIPLIER),
    )

    print("Deploying safe address: ", safe_address, ", tx: ", ethereum_tx_sent.tx_hash.hex())
    tx_receipt = w3.eth.wait_for_transaction_receipt(HexBytes(ethereum_tx_sent.tx_hash))
    if tx_receipt["status"] != 1:
        raise ValueError("Transaction failed")

    return safe_address

def generate_salt_nonce() -> int:
    return random.getrandbits(256) - 1

